import torch
import numpy as np
import math


def _flatten(sequence):
    flat = [p.contiguous().view(-1) for p in sequence]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])


def compute_cc_weights(nb_steps):
    lam = np.arange(0, nb_steps + 1, 1).reshape(-1, 1)
    lam = np.cos((lam @ lam.T) * math.pi / nb_steps)
    lam[:, 0] = .5
    lam[:, -1] = .5 * lam[:, -1]
    lam = lam * 2 / nb_steps
    W = np.arange(0, nb_steps + 1, 1).reshape(-1, 1)
    W[np.arange(1, nb_steps + 1, 2)] = 0
    W = 2 / (1 - W ** 2)
    W[0] = 1
    W[np.arange(1, nb_steps + 1, 2)] = 0
    cc_weights = torch.tensor(lam.T @ W).float()
    steps = torch.tensor(np.cos(np.arange(0, nb_steps + 1, 1).reshape(-1, 1) * math.pi / nb_steps)).float()

    return cc_weights, steps


def integrate(x0, nb_steps, step_sizes, integrand, h, compute_grad=False, x_tot=None, inv_f=False):
    #Clenshaw-Curtis Quadrature Method
    cc_weights, steps = compute_cc_weights(nb_steps)

    device = x0.get_device() if x0.is_cuda or x0.is_mps else "cpu"
    if x0.is_mps:
        device = 'mps'

    cc_weights, steps = cc_weights.to(device), steps.to(device)

    xT = x0 + nb_steps*step_sizes
    if not compute_grad:
        x0_t = x0.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        xT_t = xT.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        h_steps = h.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        steps_t = steps.unsqueeze(0).expand(x0_t.shape[0], -1, x0_t.shape[2])
        X_steps = x0_t + (xT_t-x0_t)*(steps_t + 1)/2
        X_steps = X_steps.contiguous().view(-1, x0_t.shape[2])
        h_steps = h_steps.contiguous().view(-1, h.shape[1])
        if inv_f:
            dzs = 1/integrand(X_steps, h_steps)
        else:
            dzs = integrand(X_steps, h_steps)
        dzs = dzs.view(xT_t.shape[0], nb_steps+1, -1)
        dzs = dzs*cc_weights.unsqueeze(0).expand(dzs.shape)
        z_est = dzs.sum(1)
        return z_est*(xT - x0)/2
    else:

        x0_t = x0.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        xT_t = xT.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        x_tot = x_tot * (xT - x0) / 2
        x_tot_steps = x_tot.unsqueeze(1).expand(-1, nb_steps + 1, -1) * cc_weights.unsqueeze(0).expand(x_tot.shape[0], -1, x_tot.shape[1])
        h_steps = h.unsqueeze(1).expand(-1, nb_steps + 1, -1)
        steps_t = steps.unsqueeze(0).expand(x0_t.shape[0], -1, x0_t.shape[2])
        X_steps = x0_t + (xT_t - x0_t) * (steps_t + 1) / 2
        X_steps = X_steps.contiguous().view(-1, x0_t.shape[2])
        h_steps = h_steps.contiguous().view(-1, h.shape[1])
        x_tot_steps = x_tot_steps.contiguous().view(-1, x_tot.shape[1])

        g_param, g_h = computeIntegrand(X_steps, h_steps, integrand, x_tot_steps, nb_steps+1, inv_f=inv_f)
        return g_param, g_h


def computeIntegrand(x, h, integrand, x_tot, nb_steps, inv_f=False):
    h.requires_grad_(True)
    with torch.enable_grad():
        if inv_f:
            f = 1/integrand.forward(x, h)
        else:
            f = integrand.forward(x, h)

        g_param = _flatten(torch.autograd.grad(f, integrand.parameters(), x_tot, create_graph=True, retain_graph=True))
        g_h = _flatten(torch.autograd.grad(f, h, x_tot))

    return g_param, g_h.view(int(x.shape[0]/nb_steps), nb_steps, -1).sum(1)


class ParallelNeuralIntegral(torch.autograd.Function):

    @staticmethod
    def forward(ctx, x0, x, integrand, flat_params, h, nb_steps=20, inv_f=False):
        with torch.no_grad():
            x_tot = integrate(x0, nb_steps, (x - x0)/nb_steps, integrand, h, False, inv_f=inv_f)
            # Save for backward
            ctx.integrand = integrand
            ctx.nb_steps = nb_steps
            ctx.inv_f = inv_f
            ctx.save_for_backward(x0.clone(), x.clone(), h)
        return x_tot

    @staticmethod
    def backward(ctx, grad_output):
        x0, x, h = ctx.saved_tensors
        integrand = ctx.integrand
        nb_steps = ctx.nb_steps
        inv_f = ctx.inv_f
        integrand_grad, h_grad = integrate(x0, nb_steps, x/nb_steps, integrand, h, True, grad_output, inv_f)
        x_grad = integrand(x, h)
        x0_grad = integrand(x0, h)
        # Leibniz formula
        return -x0_grad*grad_output, x_grad*grad_output, None, integrand_grad, h_grad.view(h.shape), None



"""
Implements Masked AutoEncoder for Density Estimation, by Germain et al. 2015
Re-implementation by Andrej Karpathy based on https://arxiv.org/abs/1502.03509
Modified by Antoine Wehenkel
"""

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
# ------------------------------------------------------------------------------


class MaskedLinear(nn.Linear):
    """ same as Linear except has a configurable mask on the weights """
    
    def __init__(self, in_features, out_features, bias=True):
        super().__init__(in_features, out_features, bias)        
        self.register_buffer('mask', torch.ones(out_features, in_features))
        
    def set_mask(self, mask):
        self.mask.data.copy_(torch.from_numpy(mask.astype(np.uint8).T))
        
    def forward(self, input):
        return F.linear(input, self.mask * self.weight, self.bias)


class MADE(nn.Module):
    def __init__(self, nin, hidden_sizes, nout, num_masks=1, natural_ordering=False, random=False, device="cpu"):
        """
        nin: integer; number of inputs
        hidden sizes: a list of integers; number of units in hidden layers
        nout: integer; number of outputs, which usually collectively parameterize some kind of 1D distribution
              note: if nout is e.g. 2x larger than nin (perhaps the mean and std), then the first nin
              will be all the means and the second nin will be stds. i.e. output dimensions depend on the
              same input dimensions in "chunks" and should be carefully decoded downstream appropriately.
              the output of running the tests for this file makes this a bit more clear with examples.
        num_masks: can be used to train ensemble over orderings/connections
        natural_ordering: force natural ordering of dimensions, don't use random permutations
        """
        
        super().__init__()
        self.random = random
        self.nin = nin
        self.nout = nout
        self.device = device
        self.pi = torch.tensor(math.pi).to(self.device)
        self.hidden_sizes = hidden_sizes
        assert self.nout % self.nin == 0, "nout must be integer multiple of nin"
        
        # define a simple MLP neural net
        self.net = []
        hs = [nin] + hidden_sizes + [nout]
        for h0,h1 in zip(hs, hs[1:]):
            self.net.extend([
                    MaskedLinear(h0, h1),
                    nn.ReLU(),
                ])
        self.net.pop() # pop the last ReLU for the output layer
        self.net = nn.Sequential(*self.net).to(device)
        
        # seeds for orders/connectivities of the model ensemble
        self.natural_ordering = natural_ordering
        self.num_masks = num_masks
        self.seed = 0 # for cycling through num_masks orderings
        
        self.m = {}
        self.update_masks() # builds the initial self.m connectivity
        # note, we could also precompute the masks and cache them, but this
        # could get memory expensive for large number of masks.
        
    def update_masks(self):
        if self.m and self.num_masks == 1: return # only a single seed, skip for efficiency
        L = len(self.hidden_sizes)
        
        # fetch the next seed and construct a random stream
        rng = np.random.RandomState(self.seed)
        self.seed = (self.seed + 1) % self.num_masks
        
        # sample the order of the inputs and the connectivity of all neurons
        if self.random:
            self.m[-1] = np.arange(self.nin) if self.natural_ordering else rng.permutation(self.nin)
            for l in range(L):
                self.m[l] = rng.randint(self.m[l-1].min(), self.nin-1, size=self.hidden_sizes[l])
        else:
            self.m[-1] = np.arange(self.nin)
            for l in range(L):
                self.m[l] = np.array([self.nin - 1 - (i % self.nin) for i in range(self.hidden_sizes[l])])
        
        # construct the mask matrices
        masks = [self.m[l-1][:,None] <= self.m[l][None,:] for l in range(L)]
        masks.append(self.m[L-1][:,None] < self.m[-1][None,:])
        
        # handle the case where nout = nin * k, for integer k > 1
        if self.nout > self.nin:
            k = int(self.nout / self.nin)
            # replicate the mask across the other outputs
            masks[-1] = np.concatenate([masks[-1]]*k, axis=1)
        
        # set the masks in all MaskedLinear layers
        layers = [l for l in self.net.modules() if isinstance(l, MaskedLinear)]
        for l,m in zip(layers, masks):
            l.set_mask(m)

        # map between in_d and order
        self.i_map = self.m[-1].copy()
        for k in range(len(self.m[-1])):
            self.i_map[self.m[-1][k]] = k

    
    def forward(self, x, context=None):
        if self.nout == 2:
            transf = self.net(x)
            mu, sigma = transf[:, :self.nin], transf[:, self.nin:]
            z = (x - mu) * torch.exp(-sigma)
            return z
        return self.net(x)

    def compute_ll(self, x):
        # Jac and x of MADE
        transf = self.net(x)
        mu, sigma = transf[:, :self.nin], transf[:, self.nin:]
        z = (x - mu) * torch.exp(-sigma)

        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2).sum(1)
        ll = - sigma.sum(1) + log_prob_gauss

        return ll, z

    def invert(self, z):
        if self.nin != self.nout/2:
            return None

        # We suppose a Gaussian MADE
        u = torch.zeros(z.shape)
        for d in range(self.nin):
            transf = self.forward(u)
            mu, sigma = transf[:, self.i_map[d]], transf[:, self.nin + self.i_map[d]]
            u[:, self.i_map[d]] = z[:, self.i_map[d]] * torch.exp(sigma) + mu
        return u
# ------------------------------------------------------------------------------


class ConditionnalMADE(MADE):

    def __init__(self, nin, cond_in, hidden_sizes, nout, num_masks=1, natural_ordering=False, random=False, device="cpu"):
        """
        nin: integer; number of inputs
        hidden sizes: a list of integers; number of units in hidden layers
        nout: integer; number of outputs, which usually collectively parameterize some kind of 1D distribution
              note: if nout is e.g. 2x larger than nin (perhaps the mean and std), then the first nin
              will be all the means and the second nin will be stds. i.e. output dimensions depend on the
              same input dimensions in "chunks" and should be carefully decoded downstream appropriately.
              the output of running the tests for this file makes this a bit more clear with examples.
        num_masks: can be used to train ensemble over orderings/connections
        natural_ordering: force natural ordering of dimensions, don't use random permutations
        """

        super().__init__(nin + cond_in, hidden_sizes, nout, num_masks, natural_ordering, random, device)
        self.nin_non_cond = nin
        self.cond_in = cond_in

    def forward(self, x, context):
        out = super().forward(torch.cat((context, x), 1))
        out = out.contiguous().view(x.shape[0], int(out.shape[1]/self.nin), self.nin)[:, :, self.cond_in:].contiguous().view(x.shape[0], -1)
        return out

    def computeLL(self, x, context):
        # Jac and x of MADE
        transf = self.net(torch.cat((context, x), 1))
        transf = transf.contiguous().view(x.shape[0], int(transf.shape[1] / self.nin), self.nin)[:, :, self.cond_in:].contiguous().view(x.shape[0], -1)
        mu, sigma = transf[:, :self.nin], transf[:, self.nin:]
        z = (x - mu) * torch.exp(-sigma)

        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2).sum(1)
        ll = - sigma.sum(1) + log_prob_gauss

        return ll, z

    def invert(self, z, context):
        if self.nin != self.nout / 2:
            return None

        # We suppose a Gaussian MADE
        u = torch.zeros(z.shape)
        for d in range(self.nin):
            transf = self.net(torch.cat((context, x), 1))
            mu, sigma = transf[:, self.i_map[d]], transf[:, self.nin + self.i_map[d]]
            u[:, self.i_map[d]] = z[:, self.i_map[d]] * torch.exp(sigma) + mu
        return u


if __name__ == '__main__':
    from torch.autograd import Variable
    
    # run a quick and dirty test for the autoregressive property
    D = 10
    rng = np.random.RandomState(14)
    x = (rng.rand(1, D) > 0.5).astype(np.float32)
    
    configs = [
        (D, [], D, False),                 # test various hidden sizes
        (D, [200], D, False),
        (D, [200, 220], D, False),
        (D, [200, 220, 230], D, False),
        (D, [200, 220], D, True),          # natural ordering test
        (D, [200, 220], 2*D, True),       # test nout > nin
        (D, [200, 220], 3*D, False),       # test nout > nin
    ]
    
    for nin, hiddens, nout, natural_ordering in configs:
        
        print("checking nin %d, hiddens %s, nout %d, natural %s" % 
             (nin, hiddens, nout, natural_ordering))
        model = MADE(nin, hiddens, nout, natural_ordering=natural_ordering)
        z = torch.randn(1, nin)
        model.invert(z)
        continue
        # run backpropagation for each dimension to compute what other
        # dimensions it depends on.
        res = []
        for k in range(nout):
            xtr = Variable(torch.from_numpy(x), requires_grad=True)
            xtrhat = model(xtr)
            loss = xtrhat[0,k]
            loss.backward()
            
            depends = (xtr.grad[0].numpy() != 0).astype(np.uint8)
            depends_ix = list(np.where(depends)[0])
            isok = k % nin not in depends_ix
            
            res.append((len(depends_ix), k, depends_ix, isok))
        
        # pretty print the dependencies
        res.sort()
        for nl, k, ix, isok in res:
            print("output %2d depends on inputs: %30s : %s" % (k, ix, "OK" if isok else "NOTOK"))
    



import torch
import torch.nn as nn
from .UMNNMAF import EmbeddingNetwork, UMNNMAF
import numpy as np
import math


class ListModule(object):
    def __init__(self, module, prefix, *args):
        """
        The ListModule class is a container for multiple nn.Module.
        :nn.Module module: A module to add in the list
        :string prefix:
        :list of nn.module args: Other modules to add in the list
        """
        self.module = module
        self.prefix = prefix
        self.num_module = 0
        for new_module in args:
            self.append(new_module)

    def append(self, new_module):
        if not isinstance(new_module, nn.Module):
            raise ValueError('Not a Module')
        else:
            self.module.add_module(self.prefix + str(self.num_module), new_module)
            self.num_module += 1

    def __len__(self):
        return self.num_module

    def __getitem__(self, i):
        if i < 0 or i >= self.num_module:
            raise IndexError('Out of bound')
        return getattr(self.module, self.prefix + str(i))


class UMNNMAFFlow(nn.Module):

    def __init__(self, nb_flow=1, nb_in=1, hidden_derivative=[50, 50, 50, 50], hidden_embedding=[50, 50, 50, 50],
                 embedding_s=20, nb_steps=50, act_func='ELU', solver="CC", cond_in=0, device="cpu"):
        """
        UMNNMAFFlow class is a normalizing flow made of UMNNMAF blocks.
        :int nb_flow: The number of components in the flow
        :int nb_in: The size of the input dimension (data)
        :list(int) hidden_derivative: The size of hidden layers in the integrand networks
        :list(int) hidden_embedding: The size of hidden layers in the embedding networks
        :int embedding_s: The size of the embedding
        :int nb_steps: The number of integration steps (0 for random)
        :string solver: The solver (CC or CCParallel)
        :int cond_in: The size of the conditionning variable
        :string device: The device (cpu or gpu)
        """
        super().__init__()
        self.device = device
        self.register_buffer("pi", torch.tensor(math.pi))
        self.nets = ListModule(self, "Flow")
        for i in range(nb_flow):
            auto_net = EmbeddingNetwork(nb_in, hidden_embedding, hidden_derivative, embedding_s, act_func=act_func,
                                                     device=device, cond_in=cond_in).to(device)

            model = UMNNMAF(auto_net, nb_in, nb_steps, device, solver=solver).to(device)
            self.nets.append(model)

    def to(self, device):
        for net in self.nets:
            net.to(device)
        self.device = device
        super().to(device)
        return self

    def forward(self, x, context=None):
        inv_idx = torch.arange(x.size(1) - 1, -1, -1).long()
        for net in self.nets:
            x = net.forward(x, context=context)[:, inv_idx]
        return x[:, inv_idx]

    def invert(self, z, iter=10, context=None):
        """
        From image to domain.
        :param z: A tensor of noise.
        :param iter: The number of iteration (accuracy should be around 25/100**iter
        :param context: Conditioning variable
        :return: Domain value
        """
        inv_idx = torch.arange(z.size(1) - 1, -1, -1).long()
        z = z[:, inv_idx]
        for net_i in range(len(self.nets)-1, -1, -1):
            z = self.nets[net_i].invert(z[:, inv_idx], iter, context=context)
        return z

    def compute_log_jac(self, x, context=None):
        log_jac = 0.
        inv_idx = torch.arange(x.size(1) - 1, -1, -1).long()
        for net in self.nets:
            log_jac += net.compute_log_jac(x, context=context)
            x = net.forward(x, context=context)[:, inv_idx]
        return log_jac

    def compute_log_jac_bis(self, x, context=None):
        log_jac = 0.
        inv_idx = torch.arange(x.size(1) - 1, -1, -1).long()
        for net in self.nets:
            x, l = net.compute_log_jac_bis(x, context=context)
            x = x[:, inv_idx]
            log_jac += l
        return x[:, inv_idx], log_jac

    def compute_ll(self, x, context=None):
        log_jac = 0.
        inv_idx = torch.arange(x.size(1) - 1, -1, -1).long()
        for net in self.nets:
            z = net.forward(x, context=context)[:, inv_idx]
            log_jac += net.compute_log_jac(x, context=context)
            x = z
        z = z[:, inv_idx]
        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2).sum(1)
        ll = log_jac.sum(1) + log_prob_gauss
        return ll, z

    def compute_ll_bis(self, x, context=None):
        log_jac = 0.
        inv_idx = torch.arange(x.size(1) - 1, -1, -1).long()
        for net in self.nets:
            log_jac += net.compute_log_jac(x, context=context)
            x = net.forward(x, context=context)[:, inv_idx]
        z = x[:, inv_idx]
        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2)
        ll = log_jac + log_prob_gauss
        return ll, z

    def compute_bpp(self, x, alpha=1e-6, context=None):
        d = x.shape[1]
        ll, z = self.compute_ll(x, context=context)
        bpp = -ll / (d * np.log(2)) - np.log2(1 - 2 * alpha) + 8 \
              + 1 / d * (torch.log2(torch.sigmoid(x)) + torch.log2(1 - torch.sigmoid(x))).sum(1)
        return bpp, ll, z

    def set_steps_nb(self, nb_steps):
        for net in self.nets:
            net.set_steps_nb(nb_steps)

    def compute_lipschitz(self, nb_iter=10):
        L = 1.
        for net in self.nets:
            L *= net.compute_lipschitz(nb_iter)
        return L

    def force_lipschitz(self, L=1.5):
        for net in self.nets:
            net.force_lipschitz(L)



import torch
import torch.nn as nn
from .NeuralIntegral import NeuralIntegral
from .ParallelNeuralIntegral import ParallelNeuralIntegral
import numpy as np
import math
from .made import MADE, ConditionnalMADE
class ELUPlus(nn.Module):
    def __init__(self):
        super().__init__()
        self.elu = nn.ELU()
    def forward(self, x):
        return self.elu(x) + 1.


dict_act_func = {"Sigmoid": nn.Sigmoid(), "ELU": ELUPlus()}

def _flatten(sequence):
    flat = [p.contiguous().view(-1) for p in sequence]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])


def compute_lipschitz_linear(W, nb_iter=10):
    x = torch.randn(W.shape[1], 1).to(W.device)
    for i in range(nb_iter):
        x_prev = x
        x = W.transpose(0, 1) @ (W @ x_prev)
        x = x/torch.norm(x)

    lam = (torch.norm(W.transpose(0, 1) @ (W @ x))/torch.norm(x))**.5
    return lam


class UMNNMAF(nn.Module):
    def __init__(self, net, input_size, nb_steps=100, device="cpu", solver="CC"):
        super().__init__()
        self.net = net.to(device)
        self.device = device
        self.input_size = input_size
        self.nb_steps = nb_steps
        self.cc_weights = None
        self.steps = None
        self.solver = solver
        self.register_buffer("pi", torch.tensor(math.pi))

        # Scaling could be changed to be an autoregressive network output
        self.scaling = nn.Parameter(torch.zeros(input_size, device=self.device), requires_grad=False)

    def to(self, device):
        self.device = device
        super().to(device)
        return self

    def forward(self, x, method=None, x0=None, context=None):
        x0 = x0.to(x.device) if x0 is not None else torch.zeros(x.shape).to(x.device)
        xT = x
        h = self.net.make_embeding(xT, context)
        z0 = h.view(h.shape[0], -1, x.shape[1])[:, 0, :]

        # s is a scaling factor.
        s = torch.exp(self.scaling.unsqueeze(0).expand(x.shape[0], -1))
        if self.solver == "CC":
            z = NeuralIntegral.apply(x0, x, self.net.parallel_nets, _flatten(self.net.parallel_nets.parameters()),
                                     h, self.nb_steps) + z0
        elif self.solver == "CCParallel":
            z = ParallelNeuralIntegral.apply(x0, x, self.net.parallel_nets, _flatten(self.net.parallel_nets.parameters()),
                                     h, self.nb_steps) + z0
        else:
            return None
        return s*z

    def compute_cc_weights(self, nb_steps):
        lam = np.arange(0, nb_steps + 1, 1).reshape(-1, 1)
        lam = np.cos((lam @ lam.T)*math.pi/nb_steps)
        lam[:, 0] = .5
        lam[:, -1] = .5*lam[:, -1]
        lam = lam*2/nb_steps
        W = np.arange(0, nb_steps + 1, 1).reshape(-1, 1)
        W[np.arange(1, nb_steps + 1, 2)] = 0
        W = 2/(1 - W**2)
        W[0] = 1
        W[np.arange(1, nb_steps + 1, 2)] = 0
        self.cc_weights = torch.tensor(lam.T @ W).float().to(self.device)
        self.steps = torch.tensor(np.cos(np.arange(0, nb_steps+1, 1).reshape(-1, 1) * math.pi/nb_steps)).float().to(self.device)

    def compute_log_jac(self, x, context=None):
        self.net.make_embeding(x, context)
        jac = self.net.forward(x)
        return torch.log(jac + 1e-10) + self.scaling.unsqueeze(0).expand(x.shape[0], -1)

    def compute_log_jac_bis(self, x, context=None):
        z = self.forward(x, context=context)
        jac = self.net.forward(x)
        return z, torch.log(jac + 1e-10) + self.scaling.unsqueeze(0).expand(x.shape[0], -1)

    def compute_ll(self, x, context=None):
        z = self.forward(x, context=context)
        jac = self.net.forward(x)

        z.clamp_(-10., 10.)
        log_prob_gauss = -.5 * (torch.log(self.pi * 2) + z ** 2).sum(1)
        ll = log_prob_gauss + torch.log(jac + 1e-10).sum(1) + self.scaling.unsqueeze(0).expand(x.shape[0], -1).sum(1)

        return ll, z

    def compute_ll_bis(self, x, context=None):
        z = self.forward(x, context=context)
        jac = self.net.forward(x)

        ll = torch.log(jac + 1e-10) + self.scaling.unsqueeze(0).expand(x.shape[0], -1)
        z.clamp_(-10., 10.)
        return ll, z

    def compute_bpp(self, x, alpha=1e-6, context=None):
        d = x.shape[1]
        ll, z = self.computeLL(x, context=context)
        bpp = -ll/(d*np.log(2)) - np.log2(1 - 2*alpha) + 8 \
              + 1/d * (torch.log2(torch.sigmoid(x)) + torch.log2(1 - torch.sigmoid(x))).sum(1)
        z.clamp_(-10., 10.)
        return bpp, ll, z

    def set_steps_nb(self, nb_steps):
        self.nb_steps = nb_steps

    def compute_lipschitz(self, nb_iter=10):
        return self.net.parallel_nets.computeLipshitz(nb_iter)

    def force_lipschitz(self, L=1.5):
        self.net.parallel_nets.force_lipschitz(L)

    # Kind of dichotomy with a factor 100.
    def invert(self, z, iter=10, context=None):
        nb_step = 10
        step = 1/(nb_step - 1)

        x_range = (torch.ones(z.shape[0], nb_step) * torch.arange(0, 1 + step/2, step)).permute(1, 0).to(self.device)
        z = z.unsqueeze(0).expand(nb_step, -1, -1)
        x = z.clone()
        x_inv = torch.zeros(z.shape[1], z.shape[2]).to(self.device)
        left, right = -50*torch.ones(z.shape[1], z.shape[2]).to(self.device), torch.ones(z.shape[1], z.shape[2])\
            .to(self.device)*50
        s = torch.exp(self.scaling.unsqueeze(0).unsqueeze(1).expand(x.shape[0], x.shape[1], -1))
        with torch.no_grad():
            for j in range(self.input_size):
                if j % 100 == 0:
                    print(j)

                # Compute embedding and keep only the one related to x_j
                h = self.net.make_embeding(x_inv, context)
                offset = h.view(x_inv.shape[0], -1, x_inv.shape[1])[:, 0, [j]]
                h_idx = torch.arange(j, h.shape[1], z.shape[2]).to(self.device)
                h = h[:, h_idx]

                h, offset = h.squeeze(1).unsqueeze(0).expand(nb_step, -1, -1), offset.unsqueeze(0).expand(nb_step, -1, -1)
                x0 = torch.zeros(offset.shape).view(-1, 1).to(self.device)

                derivative = lambda x, h: self.net.parallel_nets.independant_forward(torch.cat((x, h), 1))

                for i in range(iter):
                    x[:, :, j] = x_range * (right[:, j] - left[:, j]) + left[:, j]
                    # if i == 0:
                    #     print(right[:, j], left[:, j])
                    z_est = s[:, :, [j]]*(offset + ParallelNeuralIntegral.apply(x0, x[:, :, j].contiguous().view(-1, 1),
                                                                                derivative, None,
                                                                                h.contiguous().view(x0.shape[0], -1),
                                                                                self.nb_steps).contiguous().view(nb_step, -1, 1))

                    _, z_pos = torch.abs(z_est[:, :, 0] - z[:, :, j]).min(0)

                    pos_midle = z_pos + torch.arange(0, z.shape[1]).to(self.device)*nb_step
                    z_val = z_est[:, :, 0].t().contiguous().view(-1)[pos_midle]
                    x_flat = x[:, :, j].t().contiguous().view(-1)

                    mask = (z_val < z[0, :, j]).float()

                    pos_left = pos_midle - 1
                    pos_right = (pos_midle + 1) % x_flat.shape[0]

                    left[:, j] = (mask * x_flat[pos_midle] + (1 - mask) * x_flat[pos_left])
                    right[:, j] = (mask * x_flat[pos_right] + (1 - mask) * x_flat[pos_midle])
                x_inv[:, j] = x_flat[pos_midle]
        return x_inv


class IntegrandNetwork(nn.Module):
    def __init__(self, nnets, nin, hidden_sizes, nout, act_func='ELU', device="cpu"):
        super().__init__()
        self.nin = nin
        self.nnets = nnets
        self.nout = nout
        self.hidden_sizes = hidden_sizes
        self.device = device

        # define a simple MLP neural net
        self.net = []
        hs = [nin] + hidden_sizes + [nout]
        for h0, h1 in zip(hs, hs[1:]):
            self.net.extend([
                nn.Linear(h0, h1),
                nn.LeakyReLU(),
            ])
        self.net.pop()  # pop the last ReLU for the output layer
        self.net.append(dict_act_func[act_func])
        self.net = nn.Sequential(*self.net)
        self.masks = torch.eye(nnets).to(device)

    def to(self, device):
        self.device = device
        self.net.to(device)
        self.masks.to(device)
        return self

    def forward(self, x, h):
        x = torch.cat((x, h), 1)
        nb_batch, size_x = x.shape
        x_he = x.view(nb_batch, -1, self.nnets).transpose(1, 2).contiguous().view(nb_batch*self.nnets, -1)
        y = self.net(x_he).view(nb_batch, -1)
        return y

    def independant_forward(self, x):
        return self.net(x)

    def compute_lipschitz(self, nb_iter=10):
        with torch.no_grad():
            L = 1
            for layer in self.net.modules():
                if isinstance(layer, nn.Linear):
                    L *= compute_lipschitz_linear(layer.weight, nb_iter)
        return L

    def force_lipschitz(self, L=1.5):
        with torch.no_grad():
            for layer in self.net.modules():
                if isinstance(layer, nn.Linear):
                    layer.weight /= max(compute_lipschitz_linear(layer.weight, 10)/L, 1)


class EmbeddingNetwork(nn.Module):
    def __init__(self, in_d, hiddens_embedding=[50, 50, 50, 50], hiddens_integrand=[50, 50, 50, 50], out_made=1,
                 cond_in=0, act_func='ELU', device="cpu"):
        super().__init__()
        self.m_embeding = None
        self.device = device
        self.in_d = in_d
        if cond_in > 0:
            self.made = ConditionnalMADE(in_d, cond_in, hiddens_embedding, (in_d + cond_in) * (out_made), num_masks=1,
                                         natural_ordering=True).to(device)
        else:
            self.made = MADE(in_d, hiddens_embedding, in_d * (out_made), num_masks=1, natural_ordering=True).to(device)
        self.parallel_nets = IntegrandNetwork(in_d, 1 + out_made, hiddens_integrand, 1, act_func=act_func, device=device)

    def to(self, device):
        self.device = device
        self.made.to(device)
        self.parallel_nets.to(device)
        return self

    def make_embeding(self, x_made, context=None):
        self.m_embeding = self.made.forward(x_made, context)
        return self.m_embeding

    def forward(self, x_t):
        return self.parallel_nets.forward(x_t, self.m_embeding)

